import re
from xml.dom.minidom import parseString
import jenkins
import urlparse
from settings import (JENKINS_URL,
                      JENKINS_PASSWORD, JENKINS_USER,
                      PUBLIC_JENKINS_URL, URLS_TO_HIDE, logger,
                      JOBS_BLACKLISTED_FROM_MESSAGES)


def normalize_url(url):
    return urlparse.urljoin(url,
                            urlparse.urlparse(url).path.replace('//', '/'))


def is_job_published(job):
    if not job:
        logger.debug('is_job_published: no job specified')
        return False

    jenkins_server = jenkins.Jenkins(JENKINS_URL,
                                     JENKINS_USER, JENKINS_PASSWORD)
    try:
        config = jenkins_server.get_job_config(job)
        dom = parseString(config)
        elements = dom.getElementsByTagName(
            'hudson.plugins.build__publisher.BuildPublisher')
        # if the above element was found then the job is configured to be
        # published. Return true if the length of the array is > 0 and
        # False otherwise
        return elements.length > 0
    except:
        logger.info('Unable to fetch or parse job configuration for: ' +
                    job)
        return False


def hide_jenkins_url(url):
    job = None
    # Match the jenkins JOB_URL and extract the IP/hostname
    # (^http[s]?://.*) and the name of jenkins job.
    # The following format is expected
    # http://10.0.0.1:8080/job/my-jenkins-job/7/
    m = re.match('^(http[s]?://.*)/job/([^/]*)/.*$', url)
    if m:
        jenkins_ip = m.group(1)
        job = m.group(2)
    # if there was a match in the regexp above then both
    # job and jenkins_ip must be set. It is enough to check just for job
    if job and is_job_published(job):
        # if the url ends with "rebuild" its a rebuild URL. In that case
        # we don't want to replace with PUBLIC_JENKINS_URL (which is only a
        # publisher and the jobs are not running there).
        if not url.endswith('rebuild'):
            return url.replace(jenkins_ip, PUBLIC_JENKINS_URL)

    for ip, replacement in URLS_TO_HIDE:
        url = url.replace(ip, replacement)
    return url


def get_multiconfig_builds(jenkins, job_url):
    """ get list of builds for multiconfiguration jobs. The results
        is a list of dictionaries like this:
        {
          'result': 'SUCCESS',
          'output': 'http://jenkins/job/my-job/config_name/43/console'
        }"""

    job = jenkins.get_json_data(job_url + '/')
    data = []
    for run in job['runs']:
        if run['number'] == job['number']:
            result = jenkins.get_json_data(run['url'])['result']
            console_output = normalize_url(run['url'] + 'console')
            data.append({'result': result,
                         'output': console_output})
    return data


def is_matrix_job(jenkins, job_url):
    """ Check if a job on a give job_url is a matrix/multiconfiguration
    """
    #if there is a build number at the end get rid of it
    match = re.match('(^.*/)[0-9]+/?$', job_url)
    if match:
        job_url = match.group(1)
    data = jenkins.get_json_data(job_url)
    return 'activeConfigurations' in data


def parse_jenkins_url(url):
    """ Return a touple of jenkins_url, job_name and build_number given
       a jenkins url """
    build_number_match = re.search('(.*/)(.*)/([0-9]+)/?$', url)
    if build_number_match:
        build_number = int(build_number_match.group(3))
        job_name = build_number_match.group(2)
        jenkins_url = build_number_match.group(1)
        return (jenkins_url, job_name, build_number)
    else:
        return (None, None, None)


def get_executed_builds(jenkins, url):
    """ Find all executed builds no matter if the job is
    multiconfiguration/matrix or if it is triggering downstream jobs

    return value is a list of builds/dictionaries. E.g.
    {'result': 'PASSED',
     'output': 'http://jenkins:8080/job/conf1/console'}
    """

    jenkins_url, job_name, build_number = parse_jenkins_url(url)
    if not jenkins_url:
        return None

    builds = []
    if is_matrix_job(jenkins, jenkins_url + job_name):
        builds = builds + get_multiconfig_builds(jenkins, url)
    else:
        downstream_builds = get_downstream_builds(jenkins,
                                                  jenkins_url,
                                                  job_name,
                                                  build_number)
        for build in downstream_builds:
            if is_matrix_job(jenkins, build['url']):
                builds = builds + get_multiconfig_builds(jenkins,
                                                         build['url'])
            else:
                builds.append({'result': build['result'],
                               'output': normalize_url(
                                   build['url'] + '/' + 'console')})
    return builds


def get_downstream_builds(jenkins, jenkins_url, job_name, build_number,
                          depth=0):
    """For a given build find all the downstream builds and return them in
       a list where each element is of this form:
       {
       'project':
       'number':
       'url':
       'result':
       }
           """
    # this is an artifical "if" to stop the recursive calls in case something
    # goes wrong. The conditional should never really evaluate to true (unless
    # you really have job structure 10 levels deep). This mostly serves as a
    # safety net for bugs or wrong assumptions of this code against jenkins
    # behavior
    if depth > 10:
        return []
    job_url = jenkins_url + job_name
    downstream_projects = get_downstream_projects(jenkins, job_url)
    return_data = []
    json_request = '/api/json?depth=2&tree=builds[number,url,result,' + \
                   'actions[causes[upstreamBuild,upstreamProject]]]'
    for project in downstream_projects:
        data = jenkins.get_json_data(jenkins_url + project + json_request,
                                     append_api=False)
        for build in data['builds']:
            causes = []
            for action in build['actions']:
                if 'causes' in action:
                    causes = action['causes'][0]
                    break
            upstream_job = None
            upstream_build = None
            if 'upstreamProject' in causes:
                upstream_job = causes['upstreamProject']
            if 'upstreamBuild' in causes:
                upstream_build = causes['upstreamBuild']
            if (upstream_job == job_name and
                    upstream_build == build_number):
                return_data.append({'project': project,
                                   'number': build['number'],
                                   'url': normalize_url(build['url']),
                                   'result': build['result']})

    for job_build in return_data:
        new_data = get_downstream_builds(jenkins, jenkins_url,
                                         job_build['project'],
                                         job_build['number'],
                                         depth=depth + 1)
        return_data = return_data + new_data

    # filter out jobs which are on the blacklist. These are usually something
    # like 'generic-update_mp' but it can be arbitrary job. We are filtering
    # it at the very end as we might want to filter intermediate jobs too
    # (i.e. we might be only interested in results of all the downstream jobs
    # rather than the job itself)
    return_data = [p for p in return_data
                   if not p['project'] in JOBS_BLACKLISTED_FROM_MESSAGES]

    return return_data


def get_downstream_projects(jenkins, job_url):
    data = jenkins.get_json_data(job_url + '/')
    downstream_projects = []
    for project in data['downstreamProjects']:
        downstream_projects.append(project['name'])
    return downstream_projects


def get_coverity_artifacts(jenkins, url):
    ret = []
    artifacts = jenkins.get_json_data(url)['artifacts']
    for artifact in artifacts:
        path = artifact['relativePath']
        url = url.rstrip('/')
        if path.startswith('results/coverity'):
            link = "{run_url}/artifact/{artifact_path}"
            link = link.format(
                run_url=url,
                artifact_path=path)
            ret.append(normalize_url(link))
    return ret


def get_deb_artifacts(jenkins, url):
    artifacts = jenkins.get_json_data(url)['artifacts']
    for artifact in artifacts:
        path = artifact['relativePath']
        url = url.rstrip('/')
        if path.startswith('work/output'):
            link = "{run_url}/artifact/work/output/*zip*/output.zip"
            return normalize_url(link.format(run_url=url))
    return None


def get_parameter(jenkins, job_url, name, default=None):
    '''Returns the default value of a given job parameter'''
    data = jenkins.get_json_data(job_url + '/')
    for param in data['property'][1]['parameterDefinitions']:
        if param['name'] == name:
            value = param['defaultParameterValue']['value']
            logger.debug("jenkins job parameter [%s]: %s" %
                         (name, value))
            return value
    logger.debug("jenkins job parameter [%s]: not found" % (name))
    if default is None:
        raise KeyError("jenkins job parameter [%s]: not found" % (name))
    logger.debug("Returning default: %s" % (default))
    return default


def get_merge_job(jenkins, job_url):
    '''Retrieves the merge build trigger from a job'''
    merge_jobs = []
    downstream_projects = get_downstream_projects(jenkins, job_url)
    for job_name in downstream_projects:
        m = re.match('(.*-land$)', job_name)
        if m and m not in merge_jobs:
            merge_jobs.append(job_name)
    if len(merge_jobs) == 0:
        raise KeyError("No merge jobs found")
    if len(merge_jobs) > 1:
        raise KeyError("Multiple merge jobs found: %s" %
                       (', '.join(merge_jobs)))
    return merge_jobs[0]
