from launchpadlib.launchpad import Launchpad
from launchpadlib.credentials import UnencryptedFileCredentialStore
from credentials import AuthorizeRequestTokenWithConsole
from mergeproposalreview import MergeProposalReview
from settings import (ALLOWED_USERS, LAUNCHPAD_REVIEW_TYPE,
                      JENKINS_PASSWORD, JENKINS_USER, LAUNCHPAD_LOGIN,
                      CREDENTIAL_STORE_PATH,
                      LP_ENV, LP_APP, logger)
from lazr.restfulclient.errors import Unauthorized
import re
from jsonjenkins import JSONJenkins
from textwrap import dedent
import jenkinsutils
import jenkins


class LaunchpadAgent():
    _launchpad = None

    def __init__(self, launchpadlib_dir=None):
        self._launchpad = self.login(launchpadlib_dir)

    def get_launchpad(self):
        return self._launchpad

    def change_mp_status(self, mp, status, message, subject,
                         revision, vote=None):
        mp.setStatus(status=status, revid=revision)
        if vote:
            mp.createComment(review_type=LAUNCHPAD_REVIEW_TYPE,
                             subject=subject, content=message, vote=vote)
            review = MergeProposalReview(mp)
            review.unlock()
        else:
            mp.createComment(review_type=LAUNCHPAD_REVIEW_TYPE,
                             subject=subject, content=message)

    def get_merge_proposal(self, branch_name, merge_proposal_link):
        lp_branch = branch_name.replace('https://code.launchpad.net/', 'lp:')
        lp_branch = lp_branch.replace('https://code.staging.launchpad.net/',
                                      'lp://staging/')
        logger.debug('fetching branch: ' + lp_branch)
        branch = self._launchpad.branches.getByUrl(url=lp_branch)
        if not branch:
            logger.debug('branch "' + branch_name + '" doesn\'t exist')
            return None

        logger.debug('mp_link: ' + merge_proposal_link)

        for mp in branch.landing_targets:
            logger.debug('mp.web_link: ' + mp.web_link)
            if mp.web_link == merge_proposal_link:
                return mp

        return None

    def get_launchpad_user(self):
        return self._launchpad.people(LAUNCHPAD_LOGIN)

    @staticmethod
    def get_branch_from_mp(merge_proposal):
        m = re.search('(.*)\+merge/[0-9]+$', merge_proposal)
        if m:
            return m.group(1)
        return None

    @staticmethod
    def login(launchpadlib_dir=None):
        store = UnencryptedFileCredentialStore(CREDENTIAL_STORE_PATH)
        authorization_engine = AuthorizeRequestTokenWithConsole(LP_ENV, LP_APP)
        return Launchpad.login_with(LP_APP, LP_ENV,
                                    credential_store=store,
                                    authorization_engine=authorization_engine,
                                    launchpadlib_dir=launchpadlib_dir)


class UpdateMergeProposalException(Exception):
    pass


def build_state(args):
    """Return a tuple containing merge proposal state, message, and
    subject."""

    if 'merge_failed' in args and args['merge_failed']:
        return ('Needs review',
                "FAILED: Autolanding.\n" +
                "Merging failed. More details in the following jenkins job:\n",
                "Re: [Merge] autolanding failed (merge failed)")
    elif 'push_failed' in args and args['push_failed']:
        return ('Needs review',
                "FAILED: Autolanding.\n" +
                "Pushing failed. More details in the following jenkins job:\n",
                "Re: [Merge] autolanding failed (push failed)")
    elif 'landing_failed' in args and args['landing_failed']:
        return ('Needs review',
                "FAILED: Autolanding.\n" +
                "More details in the following jenkins job:\n",
                "Re: [Merge] autolanding failed")
    elif 'merged' in args and args['merged']:
        return ('Merged',
                "PASSED: Autolanding.\nBranch merged.",
                "Re: [Merge] branch merged")
    elif 'commit_message_empty' in args and args['commit_message_empty']:
        return ('Needs review',
                "FAILED: Autolanding.\n" +
                "No commit message was specified in the merge proposal. " +
                "Hit 'Add commit message' on the merge proposal web page " +
                "or follow the link below. You can approve the merge " +
                "proposal yourself to rerun.\n" +
                args['merge_proposal'] + "/+edit-commit-message\n",
                "Re: [Merge] autolanding failed",)
    elif 'unapproved_changes' in args and args['unapproved_changes']:
        return ('Needs review',
                "FAILED: Autolanding.\n" +
                "Unapproved changes made after approval.\n",
                "Re: [Merge] autolanding failed",)
    elif 'invalid_revid' in args and args['invalid_revid']:
        return ('Needs review',
                "FAILED: Autolanding.\n" +
                "Approved revid is not set in launchpad " +
                "(maybe a permission problem?).\n",
                "Re: [Merge] autolanding failed",)
    else:
        raise UpdateMergeProposalException("unrecognized parameters")


class LaunchpadVote():
    APPROVE = 'Approve'
    DISAPPROVE = 'Disapprove'
    NEEDS_FIXING = 'Needs Fixing'


class LaunchpadTrigger():
    _launchpad = None
    _launchpad_user = None
    _json_jenkins = None

    def __init__(self, launchpadlib_dir=None):
        self._launchpadAgent = LaunchpadAgent(launchpadlib_dir)
        self._launchpad = self._launchpadAgent.get_launchpad()
        self._launchpad_user = self._launchpad.people(LAUNCHPAD_LOGIN)

    def get_launchpad_user(self):
        return self._launchpad_user

    def get_branch_from_mp(self, merge_proposal):
        return self._launchpadAgent.get_branch_from_mp(merge_proposal)

    def hide_jenkins_url(self, url):
        return jenkinsutils.hide_jenkins_url(url)

    def change_mp_status(self, mp, status, message, subject,
                         revision, vote=None):
        return self._launchpadAgent.change_mp_status(mp, status, message,
                                                     subject, revision,
                                                     vote)

    def get_merge_proposal(self, branch_name, merge_proposal_link):
        return self._launchpadAgent.get_merge_proposal(
            branch_name,
            merge_proposal_link)

    def approve_mp(self, mp, revision, build_url):
        state = 'PASSED: Continuous integration, rev:' + str(revision)
        logger.debug(state)
        content = self.format_message_for_mp_update(build_url,
                                                    state + "\n")
        mp.createComment(review_type=LAUNCHPAD_REVIEW_TYPE,
                         vote=LaunchpadVote.APPROVE, subject=state,
                         content=content)
        review = MergeProposalReview(mp)
        review.unlock()

    def disapprove_mp(self, mp, revision, build_url, reason=None):
        state = "FAILED: Continuous integration, rev:{revision}".format(
            revision=revision)
        subject = state
        if reason:
            state = "{state}\n{reason}".format(state=state, reason=reason)

        logger.debug(state)
        content = self.format_message_for_mp_update(build_url, state + "\n")
        mp.createComment(review_type=LAUNCHPAD_REVIEW_TYPE,
                         vote=LaunchpadVote.NEEDS_FIXING, subject=subject,
                         content=content)
        review = MergeProposalReview(mp)
        review.unlock()

    def _get_json_jenkins(self, url):
        if not self._json_jenkins:
            self._json_jenkins = JSONJenkins(url, JENKINS_USER,
                                             JENKINS_PASSWORD)
        return self._json_jenkins

    def format_message_for_mp_update(self, build_url, message=None,
                                     include_rebuild_link=True):
        formatted_message = dedent('''\
                {message}{build_url}{executed_test_runs}
                ''')
        if include_rebuild_link:
                formatted_message = formatted_message + dedent('''\

                Click here to trigger a rebuild:
                {rebuild_url}
                ''')
        if not message:
            message = ''
        if not build_url:
            build_url = ''
            executed_test_runs_message = ''
        else:
            executed_test_runs_message = self.get_executed_test_runs_message(
                build_url)

        rebuild_url = '{build_url}/rebuild'.format(
            build_url=build_url.rstrip('/'))
        return formatted_message.format(
            message=message,
            build_url=jenkinsutils.hide_jenkins_url(build_url),
            rebuild_url=jenkinsutils.hide_jenkins_url(rebuild_url),
            executed_test_runs=executed_test_runs_message)

    def get_latest_review(self, mp):
        revision = 0
        launchpad_user = self.get_launchpad_user()
        for comment in mp.all_comments:
            if comment.author.name == launchpad_user.name:
                if comment.vote_tag == LAUNCHPAD_REVIEW_TYPE:
                    m = re.search(
                        '^(PASSED|FAILED): Continuous integration, rev:(\d+)',
                        comment.message_body)
                    if m and (int(m.group(2)) > revision):
                        revision = int(m.group(2))
        return revision

    def latest_candidate_validated(self, mp):
        latest_review = self.get_latest_review(mp)
        logger.debug('Latest_review is revision: ' + str(latest_review))
        if latest_review >= mp.source_branch.revision_count:
            logger.debug('Skipping this MP. Current revision: ' +
                         str(mp.source_branch.revision_count))
            return True
        return False

    def testing_in_progress(self, mp):
        review = MergeProposalReview(mp)
        if review.is_locked():
            logger.debug('Skipping this MP. It is curently being ' +
                         'tested by Jenkins.')
            return True
        return False

    def user_allowed_to_trigger_jobs(self, lp_user):
        allowed_people = ALLOWED_USERS
        if lp_user.name in allowed_people:
            return True
        for membership in lp_user.memberships_details:
            if membership.team.name in allowed_people:
                return True
        logger.debug('User "' + lp_user.name +
                     '" not allowed to trigger jobs')
        return False

    def is_merge_only(self, mp):
        """Determines if an mp can be merged without a build"""
        # diffstat is a dictionary of {file: changeset}
        # diffstat will be None if launchpad has not yet generated a diff
        if mp.preview_diff is None or mp.preview_diff.diffstat is None:
            raise LookupError("preview_diff of this merge proposal is None")
        diff_stat = mp.preview_diff.diffstat
        logger.debug("%s" % (diff_stat))
        if (len(diff_stat) == 1 and u'debian/changelog' in diff_stat):
            logger.debug("debian/changelog only merge")
            return True
        return False

    def add_merge_params(self, jenkins, job_url, params):
        """Populate the jenkins parameters to complete the fasttrack merge"""
        params['target_branch'] = jenkinsutils.get_parameter(
            jenkins,
            job_url,
            'target_branch')
        params['source_branch'] = params['landing_candidate']
        params['revision'] = params['candidate_revision']
        params['packaging_branch'] = jenkinsutils.get_parameter(
            jenkins,
            job_url,
            'packaging_branch')
        params['ppa_target'] = jenkinsutils.get_parameter(jenkins, job_url,
                                                          'ppa_target')
        params['distributions'] = jenkinsutils.get_parameter(jenkins,
                                                             job_url,
                                                             'distributions')
        params['buildjob_url'] = "Build was skipped, merge only"
        params['use_description_for_commit'] = jenkinsutils.get_parameter(
            jenkins,
            job_url,
            'use_description_for_commit')
        params['version_string_format'] = jenkinsutils.get_parameter(
            jenkins,
            job_url,
            'version_string_format')
        params['hooks'] = jenkinsutils.get_parameter(jenkins, job_url,
                                                     'dput_hooks', '')
        params['test_result'] = "PASSED"

    def start_jenkins_job(self, jenkins_url, jenkins_job, mp,
                          fasttrack_merge=False):
            j = jenkins.Jenkins(jenkins_url, JENKINS_USER, JENKINS_PASSWORD)
            jenkins_params = {
                'landing_candidate': 'bzr+ssh://' + LAUNCHPAD_LOGIN +
                '@bazaar.launchpad.net/' + mp.source_branch.unique_name,
                'merge_proposal': mp.web_link,
                'candidate_revision': mp.source_branch.revision_count}

            job_info = j.get_job_info(jenkins_job)
            if not job_info['buildable']:
                logger.debug('Job is not buildable (probably disabled)')
                return False

            try:
                mp.nominateReviewer(reviewer=self.get_launchpad_user(),
                                    review_type=LAUNCHPAD_REVIEW_TYPE)
            except Unauthorized:
                logger.info('You are not authorized to review this merge ' +
                            'proposal. CI process will now be stopped.')
                return False

            try:
                is_merge_only = self.is_merge_only(mp)
            except LookupError as error:
                logger.error(error.message)
                logger.debug('Skipping this merge proposal')
                return False

            if fasttrack_merge and is_merge_only:
                logger.debug('Skipping the build, merge only')
                json_jenkins = self._get_json_jenkins(jenkins_url)
                job_url = '/'.join([jenkins_url, 'job', jenkins_job]) + '/'
                logger.debug(job_url)
                merge_job = jenkinsutils.get_merge_job(json_jenkins,
                                                       job_url)
                self.add_merge_params(json_jenkins, job_url,
                                      jenkins_params)
                logger.debug('Updated params: ' + str(jenkins_params))
                jenkins_job = merge_job
            else:
                logger.debug('Doing a regular build')

            review = MergeProposalReview(mp)
            try:
                logger.debug('Starting job: ' + str(jenkins_params))
                review.lock(jenkins_job)
                j.build_job(jenkins_job, jenkins_params)
            except jenkins.JenkinsException:
                logger.debug('Starting a job failed')
                review.unlock()
                message = "ERROR: failed to start a jenkins job"
                mp.createComment(review_type=LAUNCHPAD_REVIEW_TYPE,
                                 vote='Disapprove', subject=message,
                                 content=message)

    def _get_result_line(self, result, url):
        url = url.rstrip('/')
        if result == 'SUCCESS' or result == 'UNSTABLE':
            if url.endswith('console'):
                # strip "/console"
                url = url[:-8]
        return "\n    {result}: {output}".format(
            result=result,
            output=url)

    def get_executed_test_runs_message(self, url):
        ret = "\nExecuted test runs:"
        jenkins = self._get_json_jenkins(url)
        builds = jenkinsutils.get_executed_builds(jenkins, url)
        coverity_artifacts = []
        for build in builds:
            ret += self._get_result_line(
                build['result'],
                jenkinsutils.hide_jenkins_url(build['output']))
            if build['output'].endswith('console'):
                build_url = build['output'][:-len('console')]
                artifacts = jenkinsutils.get_coverity_artifacts(jenkins,
                                                                build_url)
                debs = jenkinsutils.get_deb_artifacts(jenkins, build_url)
                if debs:
                    ret += "\n        deb: {}".format(
                        jenkinsutils.hide_jenkins_url(debs))
                coverity_artifacts += artifacts
        if coverity_artifacts:
            ret += "\nCoverity artifacts:"
            for artifact in coverity_artifacts:
                ret += "\n    {}".format(
                    jenkinsutils.hide_jenkins_url(artifact))

        return ret

    def trigger_ci_build(self, mps, jenkins_job, jenkins_url):
        result = True
        for mp in mps:

            if not self.user_allowed_to_trigger_jobs(mp.registrant):
                continue

            if self.testing_in_progress(mp):
                continue

            if self.latest_candidate_validated(mp):
                continue

            ret = self.start_jenkins_job(jenkins_url, jenkins_job, mp)
            result = ret and result
        return result

    def unapproved_prerequisite_exists(self, mp):
        prereq = mp.prerequisite_branch
        target_branch = mp.target_branch
        if prereq:
            #x is a merge-proposal object
            merges = [x for x in prereq.landing_targets
                      if x.target_branch.web_link == target_branch.web_link and
                      x.queue_status != u'Superseded']
            if len(merges) == 0:
                logger.debug('No proposals of prerequisite branch exists')
                return True
            else:
                for merge in merges:
                    #lets make our life simpler here and lets assume
                    #all the MP needs to be merged even though in reality
                    #their target branch might differ
                    if merge.queue_status != u'Merged':
                        logger.debug('Prerequisite not yet merged')
                        return True
        #TODO
        # skip also branches with changes made after approval
        return False

    def is_commit_message_set(self, mp, jenkins_job, jenkins_url):
        jenkins = self._get_json_jenkins(jenkins_url)
        job_url = jenkins_url + '/job/' + jenkins_job + '/'
        use_description = False
        try:
            if jenkinsutils.get_parameter(jenkins, job_url,
                                          'use_description_for_commit'):
                use_description = True
        except KeyError:
            logger.debug("Unable to get \"use_description_for_commit\" " +
                         "configuration for %s" % (jenkins_job))
        logger.debug('Job is configured to fallback to description: %s' %
                     (use_description))
        return bool(self.get_commit_message(mp, use_description))

    def trigger_al_build(self, mps, jenkins_job, jenkins_url, fasttrack_merge):
        result = True
        for mp in mps:
            #ignore this merge proposal if both registrant and reviewer
            #are not trusted (aka belong to canonical)
            #check if mp.reviewer actually exists just to be sure (it should
            # as we are autolanding and thus the MP needs to be approved)
            if mp.reviewer is None:
                continue
            if not (self.user_allowed_to_trigger_jobs(mp.registrant) or
                    self.user_allowed_to_trigger_jobs(mp.reviewer)):
                logger.debug('Skiping this merge proposal as neither ' +
                             'registrant nor reviewer is allowed to ' +
                             'trigger jenkins jobs')
                continue

            if not self.is_commit_message_set(mp, jenkins_job, jenkins_url):
                logger.debug('No commit message was set. Interrupting ' +
                             'autolanding for "%s".' % (mp.web_link))
                mp_state, message, subject = build_state({
                    'commit_message_empty': True,
                    'merge_proposal': mp.web_link})
                self.change_mp_status(mp, mp_state, message, subject,
                                      mp.source_branch.revision_count,
                                      LaunchpadVote.NEEDS_FIXING)
                continue

            if self.testing_in_progress(mp):
                continue

            if self.unapproved_prerequisite_exists(mp):
                logger.debug('Unapproved prerequisite exists for ' + str(mp))
                continue

            logger.debug('Going to trigger merge: ' + str(mp.source_branch))

            ret = self.start_jenkins_job(jenkins_url, jenkins_job, mp,
                                         fasttrack_merge)
            result = ret and result
        return result

    def get_merge_proposals(self, branch_name, state):
        branch = self._launchpad.branches.getByUrl(url=branch_name)
        if not branch:
            logger.debug('Branch "' + branch_name + '" doesn\'t exist')
            return []
        mps = []
        for mp in branch.landing_candidates:
            if mp.queue_status in state:
                mps.append(mp)
        return mps

    def get_reviews(self, mp):
        reviews = []
        for vote in mp.votes:
            if vote.comment:
                reviews.append('%s;%s' % (vote.reviewer.display_name,
                                          vote.comment.vote))
        if len(reviews) == 0:
            return None

        return reviews

    def get_commit_message(self, mp, use_description):
        message = None
        if mp.commit_message:
            message = mp.commit_message
        elif use_description and mp.description:
            message = mp.description
        return message

    def close_bugs(self, mp):
        for b in mp.source_branch.linked_bugs_collection:
            for task in b.bug_tasks_collection:
                if task.target.web_link == mp.target_branch.project.web_link:
                    task.status = 'Fix Committed'
                    task.lp_save()
                    message = 'Fix committed into {target} at revision ' + \
                              '{revision}, scheduled for release in ' + \
                              '{project}, milestone {milestone}'
                    mp.lp_refresh()
                    project = mp.target_branch.project
                    milestone_name = 'Unknown'
                    active_milestones = project.active_milestones_collection
                    for active_milestone in active_milestones:
                        series = active_milestone.series_target
                        if series.branch_link == mp.target_branch.self_link:
                            milestone_name = active_milestone.name
                    message = message.format(
                        target=mp.target_branch.display_name,
                        revision=mp.merged_revno,
                        project=project.name,
                        milestone=milestone_name)
                    b.newMessage(content=message)
