import unittest
import launchpad
from textwrap import dedent
from settings import LAUNCHPAD_LOGIN, LAUNCHPAD_REVIEW_TYPE
from mock import MagicMock, patch
from jenkins import JenkinsException
from tests import JenkinsJSONData
from testscenarios import TestWithScenarios
from lazr.restfulclient.errors import Unauthorized


class LaunchpadTriggerInstantiation(unittest.TestCase):
    def test_instatiation(self):
        launchpadAgent = MagicMock(name="launchpadAgentMock")
        launchpadAgent.login = MagicMock(return_value=MagicMock())
        with patch('launchpad.LaunchpadAgent', new=launchpadAgent):
            l = launchpad.LaunchpadTrigger()
            l._launchpad.people.assert_called_once_with(LAUNCHPAD_LOGIN)


class LaunchpadTriggerTestCase(unittest.TestCase):
    launchpadTrigger = None
    jenkinsUserName = 'jenkins-user'

    def setUp(self):
        launchpadAgent = MagicMock(name="launchpadAgentMock")
        launchpadAgent.login = MagicMock(return_value=MagicMock())
        user = MagicMock()
        user.name = self.jenkinsUserName
        with patch('launchpad.LaunchpadAgent', new=launchpadAgent):
            self.launchpadTrigger = launchpad.LaunchpadTrigger()
            self.launchpadTrigger._launchpad_user = user
        self.launchpadTrigger.get_executed_test_runs_message = lambda self: ''
        self.launchpadTrigger.get_coverity_artifacts = lambda self: ''

    def test_new_mp_candidate_revision_in_queue(self):
            #60 is the revision which was validated as last
            self.launchpadTrigger.get_latest_review = MagicMock(
                return_value=60)
            mp = MagicMock()
            #this is the current branch revision
            mp.source_branch.revision_count = 61
            #latest_candidate_validated should return False as there is a new
            #revision in the queue (61)
            self.assertFalse(
                self.launchpadTrigger.latest_candidate_validated(mp))

    def test_mp_already_validated(self):
            #60 is the revision which was validated as last
            self.launchpadTrigger.get_latest_review = MagicMock(
                return_value=60)
            mp = MagicMock()
            #this is the current branch revision
            mp.source_branch.revision_count = 60
            #latest_candidate_validated should return True. This rev was
            #already validated
            self.assertTrue(
                self.launchpadTrigger.latest_candidate_validated(mp))

    def test_get_latest_review_no_comments(self):
        mp = MagicMock()
        mp.all_comments = []
        self.assertEqual(self.launchpadTrigger.get_latest_review(mp), 0)

    def test_get_latest_review_unrelated_comment(self):
        mp = MagicMock()
        comment = MagicMock()
        comment.vote_tag = LAUNCHPAD_REVIEW_TYPE
        comment.message_body = 'unrelated comment'
        mp.all_comments = [comment]
        self.assertEqual(self.launchpadTrigger.get_latest_review(mp), 0)

    def test_get_latest_review_which_passed(self):
        mp = MagicMock()
        comment = MagicMock()
        comment.author.name = self.jenkinsUserName
        comment.vote_tag = LAUNCHPAD_REVIEW_TYPE
        comment.message_body = "PASSED: Continuous integration, rev:47\n" + \
                               "URL:..."
        mp.all_comments = [comment]
        self.assertEqual(self.launchpadTrigger.get_latest_review(mp), 47)

    def test_get_latest_review_which_failed(self):
        mp = MagicMock()
        comment = MagicMock()
        comment.author.name = self.jenkinsUserName
        comment.vote_tag = LAUNCHPAD_REVIEW_TYPE
        comment.message_body = "FAILED: Continuous integration, rev:47\n" + \
                               "URL:..."
        mp.all_comments = [comment]
        self.assertEqual(self.launchpadTrigger.get_latest_review(mp), 47)

    def test_testing_not_in_progress_approved(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        review = launchpad.MergeProposalReview(mp)
        review.unlock()
        self.assertFalse(self.launchpadTrigger.testing_in_progress(mp))

    def test_testing_not_in_progress_abstain(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        review = launchpad.MergeProposalReview(mp)
        review.lock('my-job')
        self.assertTrue(self.launchpadTrigger.testing_in_progress(mp))
        self.assertEquals(review.get_jenkins_job(), 'my-job')

    def test_user_not_allowed_to_trigger_job(self):
        registrant = MagicMock()
        registrant.name = 'random-user'
        membership = MagicMock()
        membership.team.name = 'random-team'
        registrant.memberships_details = [membership]
        result = self.launchpadTrigger.user_allowed_to_trigger_jobs(registrant)
        self.assertFalse(result)

    def test_user_member_of_allowed_team(self):
        registrant = MagicMock()
        registrant.name = 'random-user'
        membership = MagicMock()
        membership.team.name = 'ps-jenkins'
        registrant.memberships_details = [membership]
        result = self.launchpadTrigger.user_allowed_to_trigger_jobs(registrant)
        self.assertTrue(result)

    def test_user_allowed_to_trigger(self):
        registrant = MagicMock()
        registrant.name = 'ps-jenkins'
        membership = MagicMock()
        membership.team.name = ''
        registrant.memberships_details = [membership]
        result = self.launchpadTrigger.user_allowed_to_trigger_jobs(registrant)
        self.assertTrue(result)

    def test_start_jenkins_job(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': True})
        jenkins = MagicMock()
        jenkins.Jenkins = MagicMock(return_value=j)
        with patch('launchpad.jenkins', new=jenkins):
            self.launchpadTrigger.start_jenkins_job('url', 'jenkins_job', mp)
            self.assertEqual(j.build_job.call_count, 1)
            self.assertEqual(mp.nominateReviewer.call_count, 1)

    def test_start_jenkins_job_while_diff_is_not_yet_generated(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = None
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': True})
        jenkins = MagicMock()
        jenkins.Jenkins = MagicMock(return_value=j)

        with patch('launchpad.jenkins', new=jenkins):
            ret = self.launchpadTrigger.start_jenkins_job('url',
                                                          'jenkins_job',
                                                          mp)
            self.assertFalse(ret)
            self.assertEqual(j.build_job.call_count, 0)

    def test_start_jenkins_job_and_raise_exception(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': True})
        j.build_job.side_effect = JenkinsException("jenkinsException")
        jenkinsInstance = MagicMock()
        jenkinsInstance.Jenkins = MagicMock(return_value=j)
        jenkinsInstance.JenkinsException = JenkinsException
        with patch('launchpad.jenkins', new=jenkinsInstance):
            self.launchpadTrigger.start_jenkins_job('url', 'jenkins_job', mp)
            self.assertEqual(mp.createComment.call_count, 1)

    def test_start_jenkins_job_and_raise_unauthorized(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.nominateReviewer.side_effect = Unauthorized("unauthorized", '')
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': True})
        jenkinsInstance = MagicMock()
        jenkinsInstance.Jenkins = MagicMock(return_value=j)
        with patch('launchpad.jenkins', new=jenkinsInstance):
            ret = self.launchpadTrigger.start_jenkins_job('url',
                                                          'jenkins_job',
                                                          mp)
            self.assertEqual(mp.createComment.call_count, 0)
            self.assertEqual(ret, False)

    def test_start_nonbuildable_jenkins_job(self):
        mp = MagicMock()
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': False})
        jenkins = MagicMock()
        jenkins.Jenkins = MagicMock(return_value=j)
        with patch('launchpad.jenkins', new=jenkins):
            self.assertEqual(
                self.launchpadTrigger.start_jenkins_job('url',
                                                        'jenkins_job',
                                                        mp),
                False)
            self.assertEqual(j.build_job.call_count, 0)
            self.assertEqual(mp.nominateReviewer.call_count, 0)
            self.assertEqual(mp.createComment.call_count, 0)

    def test_start_jenkins_job_with_fasttrack_merge(self):
        """Verifies that the merge job is called with merge only input"""
        json_jenkins = MagicMock()
        json_jenkins.get_json_data = MagicMock(
            return_value={'downstreamProjects':
                          [{'name': 'generic-cleanup'},
                           {'name': 'generic-land'}]})
        self.launchpadTrigger._get_json_jenkins = MagicMock(
            return_value=json_jenkins)
        self.launchpadTrigger.add_merge_params = MagicMock()
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        diff_stat = {u'debian/changelog': 'diff text'}
        mp.preview_diff.diffstat = diff_stat
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': True})
        jenkins = MagicMock()
        jenkins.Jenkins = MagicMock(return_value=j)
        with patch('launchpad.jenkins', new=jenkins):
            self.launchpadTrigger.start_jenkins_job('url', 'jenkins_job', mp,
                                                    True)
            self.assertEqual(j.build_job.call_count, 1)
            self.assertEqual(j.build_job.call_args[0][0], 'generic-land')

    def test_start_jenkins_job_with_fasttrack_merge_needs_build(self):
        """Verifies that the jenkins job is called when build needed"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        self.launchpadTrigger.is_merge_only = MagicMock(return_value=False)
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': True})
        jenkins = MagicMock()
        jenkins.Jenkins = MagicMock(return_value=j)
        with patch('launchpad.jenkins', new=jenkins):
            self.launchpadTrigger.start_jenkins_job('url', 'jenkins_job', mp,
                                                    True)
            self.assertEqual(j.build_job.call_count, 1)
            self.assertEqual(j.build_job.call_args[0][0], 'jenkins_job')

    def test_is_merge_only(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        diff_stat = {u'debian/changelog': 'diff text'}
        mp.preview_diff.diffstat = diff_stat
        self.assertTrue(self.launchpadTrigger.is_merge_only(mp))

    def test_is_merge_only_not_changelog(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        diff_stat = {u'debian/rules': 'diff text'}
        mp.preview_diff.diffstat = diff_stat
        self.assertFalse(self.launchpadTrigger.is_merge_only(mp))

    def test_is_merge_only_multiple_files(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        diff_stat = {u'debian/changelog': 'diff text',
                     u'debian/rules': 'diff text'}
        mp.preview_diff.diffstat = diff_stat
        self.assertFalse(self.launchpadTrigger.is_merge_only(mp))

    def test_is_merge_only_no_preview_diff(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = None
        with self.assertRaises(LookupError):
            self.launchpadTrigger.is_merge_only(mp)

    def test_is_merge_only_no_diffstat(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        mp.preview_diff.diffstat = None
        with self.assertRaises(LookupError):
            self.launchpadTrigger.is_merge_only(mp)

    def test_add_merge_params(self):
        """Verifies add_merge_params adds the needed fasttrack parameters"""
        params = {'landing_candidate': 'branch-name',
                  'merge_proposal': 'http://launchpad/mp',
                  'candidate_revision': 100}
        get_parameters = MagicMock()
        get_parameters.side_effect = ['lp:target', '', 'ppa:project/ppa',
                                      'quantal,raring', True, 'format',
                                      'H00hook']
        with patch('jenkinsutils.get_parameter', get_parameters):
            self.launchpadTrigger.add_merge_params(None, None, params)
            self.assertEqual(params['target_branch'], 'lp:target')
            self.assertEqual(params['source_branch'], 'branch-name')
            self.assertEqual(params['revision'], 100)
            self.assertEqual(params['packaging_branch'], '')
            self.assertEqual(params['ppa_target'], 'ppa:project/ppa')
            self.assertEqual(params['distributions'], 'quantal,raring')
            self.assertEqual(params['buildjob_url'],
                             'Build was skipped, merge only')
            self.assertEqual(params['use_description_for_commit'], True)
            self.assertEqual(params['version_string_format'], 'format')
            self.assertEqual(params['hooks'], 'H00hook')

    def test_unapproved_prerequisite_no_prerequisite(self):
        mp = MagicMock()
        mp.prerequisite_branch = None
        self.assertFalse(
            self.launchpadTrigger.unapproved_prerequisite_exists(mp))

    def test_unapproved_prerequisite_too_many_proposals(self):
        mp = MagicMock()
        mp.target_branch = MagicMock()
        mp.target_branch.web_link = 'url'
        prereq = MagicMock()
        prereq.landing_targets = [MagicMock(), MagicMock(), MagicMock()]
        for t in prereq.landing_targets:
                t.target_branch.web_link = 'url'
                t.queue_status = 'Approved'
        mp.prerequisite_branch = prereq
        self.assertTrue(
            self.launchpadTrigger.unapproved_prerequisite_exists(mp))

    def test_unapproved_prerequisite_no_proposal_for_prerequisite(self):
        mp = MagicMock()
        prereq = MagicMock()
        prereq.landing_targets = []
        mp.prerequisite_branch = prereq
        self.assertTrue(
            self.launchpadTrigger.unapproved_prerequisite_exists(mp))

    def test_unapproved_prerequisite_not_merged_yet(self):
        mp = MagicMock()
        mp.target_branch = MagicMock()
        mp.target_branch.web_link = 'url'
        prereq = MagicMock()
        prereq.landing_targets = [MagicMock()]
        for t in prereq.landing_targets:
                t.target_branch.web_link = 'url'
                t.queue_status = 'Approved'
        mp.prerequisite_branch = prereq
        self.assertTrue(
            self.launchpadTrigger.unapproved_prerequisite_exists(mp))

    def test_unapproved_prerequisite_merged(self):
        mp = MagicMock()
        mp.target_branch = MagicMock()
        mp.target_branch.web_link = 'url'
        prereq = MagicMock()
        prereq.landing_targets = [MagicMock()]
        for t in prereq.landing_targets:
                t.target_branch.web_link = 'url'
                t.queue_status = 'Merged'
        mp.prerequisite_branch = prereq
        self.assertFalse(
            self.launchpadTrigger.unapproved_prerequisite_exists(mp))

    def test_get_reviews_no_reviews(self):
        mp = MagicMock()
        mp.votes = []
        self.assertTrue(self.launchpadTrigger.get_reviews(mp) is None)

    def test_get_reviews(self):
        mp = MagicMock()
        mp.votes = [MagicMock(), MagicMock(), MagicMock()]
        for vote in mp.votes:
            vote.reviewer.display_name = 'jenkins'
            vote.comment.vote = 'Approve'
        reviews = self.launchpadTrigger.get_reviews(mp)
        for r in reviews:
            self.assertTrue(r == 'jenkins;Approve')

    def test_close_bugs(self):
        mp = MagicMock()
        mp.target_branch.project.web_link = 'http://launchpad/jenkins'
        mp.target_branch.project.name = 'project'
        mp.source_branch.linked_bugs_collection = [MagicMock(), MagicMock()]
        mp.target_branch.display_name = 'lp:project'
        mp.merged_revno = '12'
        mp.target_branch.self_link = 'http://project/project'
        milestone = MagicMock()
        milestone.series_target.branch_link = mp.target_branch.self_link
        milestone.name = '5.0'
        mp.target_branch.project.active_milestones_collection = [milestone]
        for b in mp.source_branch.linked_bugs_collection:
            b.bug_tasks_collection = [MagicMock(), MagicMock(), MagicMock()]
            bug_task = b.bug_tasks_collection[1]
            bug_task.target.web_link = mp.target_branch.project.web_link

        self.launchpadTrigger.close_bugs(mp)
        for b in mp.source_branch.linked_bugs_collection:
            b.bug_tasks_collection[1].lp_save.assert_called_once_with()
            b.newMessage.assert_called_once_with(
                content='Fix committed into lp:project at revision ' +
                '12, scheduled for release in project, milestone 5.0')

    def test_approve_mp(self):
        mp = MagicMock()
        mp.web_link = 'http://mp'
        self.launchpadTrigger.approve_mp(mp, '1', 'http://url')
        subject = 'PASSED: Continuous integration, rev:1'
        content = dedent('''\
                            {subject}
                            http://url

                            Click here to trigger a rebuild:
                            http://url/rebuild
                            ''').format(subject=subject)
        mp.createComment.assert_called_once_with(
            review_type=LAUNCHPAD_REVIEW_TYPE,
            vote=launchpad.LaunchpadVote.APPROVE,
            subject=subject,
            content=content)

    def test_disapprove_mp(self):
        mp = MagicMock()
        mp.web_link = 'http://mp'
        self.launchpadTrigger.disapprove_mp(mp, '1', 'http://url')
        subject = 'FAILED: Continuous integration, rev:1'
        content = dedent('''\
                            {subject}
                            http://url

                            Click here to trigger a rebuild:
                            http://url/rebuild
                            ''').format(subject=subject)
        mp.createComment.assert_called_once_with(
            review_type=LAUNCHPAD_REVIEW_TYPE,
            vote=launchpad.LaunchpadVote.NEEDS_FIXING,
            subject=subject,
            content=content)

    def test_disapprove_mp_with_reason(self):
        mp = MagicMock()
        mp.web_link = 'http://mp'
        subject = 'FAILED: Continuous integration, rev:1'
        reason = "No commit message set\nDo something\n"
        content = dedent('''\
                            {subject}
                            {reason}
                            http://url

                            Click here to trigger a rebuild:
                            http://url/rebuild
                            ''').format(subject=subject, reason=reason)

        self.launchpadTrigger.disapprove_mp(mp,
                                            '1',
                                            'http://url',
                                            reason=reason)
        mp.createComment.assert_called_once_with(
            review_type=LAUNCHPAD_REVIEW_TYPE,
            vote=launchpad.LaunchpadVote.NEEDS_FIXING,
            subject=subject,
            content=content)


class TestLaunchpadTrigger(unittest.TestCase):
    def setUp(self):
        launchpadAgent = MagicMock(name="launchpadAgentMock")
        launchpadAgent.login = MagicMock(return_value=MagicMock())
        with patch('launchpad.LaunchpadAgent', new=launchpadAgent):
            self.launchpadTrigger = launchpad.LaunchpadTrigger()


class TestGetResultLine(TestWithScenarios, TestLaunchpadTrigger):
    scenarios = [
        ('success',
         {'result': 'SUCCESS',
          'url': 'http://10.0.0.1:8080/job/my-job/13/console',
          'expected': "\n    SUCCESS: http://10.0.0.1:8080/job/my-job/13"}),
        ('success_with_slash',
         {'result': 'SUCCESS',
          'url': 'http://10.0.0.1:8080/job/my-job/13/console/',
          'expected': "\n    SUCCESS: http://10.0.0.1:8080/job/my-job/13"}),
        ('unstable',
         {'result': 'UNSTABLE',
          'url': 'http://10.0.0.1:8080/job/my-job/13/console',
          'expected': "\n    UNSTABLE: http://10.0.0.1:8080/job/my-job/13"}),
        ('failed',
         {'result': 'FAILURE',
          'url': 'http://10.0.0.1:8080/job/my-job/13/console',
          'expected':
          "\n    FAILURE: http://10.0.0.1:8080/job/my-job/13/console"}),
        ('none',
         {'result': None,
          'url': 'http://10.0.0.1:8080/job/my-job/13/console',
          'expected':
          "\n    None: http://10.0.0.1:8080/job/my-job/13/console"}),
    ]

    def test_get_result_line(self):
        result = self.launchpadTrigger._get_result_line(self.result, self.url)
        self.assertEqual(result, self.expected)


class TestGetCommitMessage(TestWithScenarios, TestLaunchpadTrigger):
    scenarios = [
        ('no_commit_message',
         {'commit_message': None,
          'description': 'description',
          'use_description': False,
          'expected': None}),
        ('no_commit_message_fallback_to_description',
         {'commit_message': None,
          'description': 'description',
          'use_description': True,
          'expected': 'description'}),
        ('commit_message_and_description',
         {'commit_message': 'commit_message',
          'description': 'description',
          'use_description': True,
          'expected': 'commit_message'}),
        ('no_commit_message_and_no_description',
         {'commit_message': '',
          'description': '',
          'use_description': True,
          'expected': None})]

    def test_get_commit_message(self):
        mp = MagicMock()
        mp.commit_message = self.commit_message
        mp.description = self.description
        use_description = self.use_description

        message = self.launchpadTrigger.get_commit_message(mp, use_description)
        self.assertEqual(message, self.expected)


class LaunchpadTriggerCIWithWrongUser(TestLaunchpadTrigger):
    def setUp(self):
        super(LaunchpadTriggerCIWithWrongUser, self).setUp()
        self.launchpadTrigger.user_allowed_to_trigger_jobs = MagicMock(
            return_value=False)
        self.launchpadTrigger.start_jenkins_job = MagicMock()

    def test_trigger_ci_with_wrong_user(self):
        self.launchpadTrigger.trigger_ci_build([MagicMock()], 'job', 'url')
        self.assertEqual(self.launchpadTrigger.start_jenkins_job.call_count,
                         0)

    def test_trigger_ci_with_no_mps(self):
        self.launchpadTrigger.trigger_ci_build([], 'job', 'url')
        self.assertEqual(self.launchpadTrigger.start_jenkins_job.call_count,
                         0)


class LaunchpadTriggerCIWhileTestingInProgress(TestLaunchpadTrigger):
    def setUp(self):
        super(LaunchpadTriggerCIWhileTestingInProgress, self).setUp()
        self.launchpadTrigger.user_allowed_to_trigger_jobs = MagicMock()
        self.launchpadTrigger.testing_in_progress = MagicMock(
            return_value=True)
        self.launchpadTrigger.start_jenkins_job = MagicMock()

    def test_trigger_ci_while_testing_in_progress(self):
        self.launchpadTrigger.trigger_ci_build([MagicMock()], 'job', 'url')
        self.assertEqual(self.launchpadTrigger.start_jenkins_job.call_count,
                         0)


class LaunchpadTriggerCIWhileRevisionValidated(TestLaunchpadTrigger):
    def setUp(self):
        super(LaunchpadTriggerCIWhileRevisionValidated, self).setUp()
        self.launchpadTrigger.user_allowed_to_trigger_jobs = MagicMock()
        self.launchpadTrigger.testing_in_progress = MagicMock(
            return_value=False)
        self.launchpadTrigger.latest_candidate_validated = MagicMock(
            return_value=True)
        self.launchpadTrigger.start_jenkins_job = MagicMock()

    def test_trigger_ci_while_Revision_validated(self):
        self.launchpadTrigger.trigger_ci_build([MagicMock()], 'job', 'url')
        self.assertEqual(self.launchpadTrigger.start_jenkins_job.call_count,
                         0)


class LaunchpadTriggerCIW(TestLaunchpadTrigger):
    def setUp(self):
        super(LaunchpadTriggerCIW, self).setUp()
        self.launchpadTrigger.user_allowed_to_trigger_jobs = MagicMock()
        self.launchpadTrigger.testing_in_progress = MagicMock(
            return_value=False)
        self.launchpadTrigger.latest_candidate_validated = MagicMock(
            return_value=False)
        self.launchpadTrigger.start_jenkins_job = MagicMock()

    def test_trigger_ci_while_Revision_validated(self):
        self.launchpadTrigger.trigger_ci_build([MagicMock()], 'job', 'url')
        self.assertEqual(self.launchpadTrigger.start_jenkins_job.call_count,
                         1)


class LaunchpadTriggerAutolandingWithoutPermissions(TestLaunchpadTrigger):
    def setUp(self):
        super(LaunchpadTriggerAutolandingWithoutPermissions, self).setUp()
        self.launchpadTrigger.user_allowed_to_trigger_jobs = MagicMock(
            return_value=False)
        self.launchpadTrigger.start_jenkins_job = MagicMock()

    def test_trigger_al_without_permissions(self):
        self.launchpadTrigger.trigger_al_build(
            [MagicMock(), MagicMock(), MagicMock()], 'job', 'url', None)
        self.assertEqual(self.launchpadTrigger.start_jenkins_job.call_count,
                         0)


class LaunchpadTriggerAutolandingWithTrustedReviewer(TestLaunchpadTrigger):
    def setUp(self):
        super(LaunchpadTriggerAutolandingWithTrustedReviewer, self).setUp()
        self.launchpadTrigger.user_allowed_to_trigger_jobs = lambda x: x
        self.launchpadTrigger.is_commit_message_set = MagicMock(
            return_value=True)

        self.launchpadTrigger.start_jenkins_job = MagicMock()
        #return True so we stop the flow; We just want to test the permissions
        #above and not the whole trigger_al_build method
        test_in_progress = MagicMock(return_value=True)
        self.launchpadTrigger.testing_in_progress = test_in_progress

    def test_trigger_al_with_trusted_reviewer(self):
        mp = MagicMock()
        #user_allowed_to_trigger_jobs now returns what it gets as an argument
        #(see the lambda in constructor)
        #We want to return False for registrant (i.e. person who create MP is
        #not allowed to trigger jobs in our infrastructure)
        #We want to return True for reviewer (i.e. the person who reviewed is
        #trusted and we can run the code on his behalf)
        mp.registrant = False
        mp.reviewer = True
        self.launchpadTrigger.trigger_al_build([mp], 'job', 'url', None)
        self.assertEqual(self.launchpadTrigger.testing_in_progress.call_count,
                         1)

    def test_trigger_al_with_trusted_registrant(self):
        mp = MagicMock()
        #user_allowed_to_trigger_jobs now returns what it gets as an argument
        #(see the lambda in constructor)
        #in this test-case even if the reviewer is not trusted we trust the
        #registrant so we are good to go
        mp.registrant = True
        mp.reviewer = False
        self.launchpadTrigger.trigger_al_build([mp], 'job', 'url', None)
        self.assertEqual(self.launchpadTrigger.testing_in_progress.call_count,
                         1)

    def test_trigger_al_with_untrusted_people(self):
        mp = MagicMock()
        #user_allowed_to_trigger_jobs now returns what it gets as an argument
        #(see the lambda in constructor)
        #in this case none of the persons are trusted and therefore
        #we should stop and testing_in_progress shouldn't be executed at all
        mp.registrant = False
        mp.reviewer = False
        self.launchpadTrigger.trigger_al_build([mp], 'job', 'url', None)
        self.assertEqual(self.launchpadTrigger.testing_in_progress.call_count,
                         0)


class LaunchpadTriggerAutolandingWhileTestingInProgress(TestLaunchpadTrigger):
    def setUp(self):
        super(LaunchpadTriggerAutolandingWhileTestingInProgress, self).setUp()
        self.launchpadTrigger.user_allowed_to_trigger_jobs = MagicMock(
            return_value=True)
        self.launchpadTrigger.is_commit_message_set = MagicMock(
            return_value=False)
        self.launchpadTrigger.start_jenkins_job = MagicMock()

    def test_trigger_al_while_testing_in_progress(self):
        self.launchpadTrigger.trigger_al_build(
            [MagicMock(), MagicMock(), MagicMock()], 'job', 'url', None)
        self.assertEqual(self.launchpadTrigger.start_jenkins_job.call_count,
                         0)


class LaunchpadTriggerAutolandingWithNoCommitMessage(TestLaunchpadTrigger):
    def setUp(self):
        super(LaunchpadTriggerAutolandingWithNoCommitMessage, self).setUp()
        self.launchpadTrigger.change_mp_status = MagicMock()
        self.launchpadTrigger.user_allowed_to_trigger_jobs = MagicMock(
            return_value=True)
        self.launchpadTrigger.testing_in_progress = MagicMock(
            return_value=False)
        self.launchpadTrigger.is_commit_message_set = MagicMock(
            return_value=False)
        self.launchpadTrigger.start_jenkins_job = MagicMock()

    def test_trigger_al_while_testing_in_progress(self):
        self.launchpadTrigger.trigger_al_build(
            [MagicMock(), MagicMock(), MagicMock()], 'job', 'url', None)
        self.assertEqual(self.launchpadTrigger.change_mp_status.call_count,
                         3)


class AutolandingWithUnapprovedPrerequisite(TestLaunchpadTrigger):
    def setUp(self):
        super(AutolandingWithUnapprovedPrerequisite, self).setUp()
        self.launchpadTrigger.user_allowed_to_trigger_jobs = MagicMock(
            return_value=True)
        self.launchpadTrigger.testing_in_progress = MagicMock(
            return_value=False)
        self.launchpadTrigger.unapproved_prerequisite_exists = MagicMock(
            return_value=True)
        self.launchpadTrigger.is_commit_message_set = MagicMock(
            return_value=True)

        self.launchpadTrigger.start_jenkins_job = MagicMock()

    def test_trigger_al_with_unapproved_prerequisite(self):
        self.launchpadTrigger.trigger_al_build(
            [MagicMock(), MagicMock(), MagicMock()], 'job', 'url', None)
        self.assertEqual(self.launchpadTrigger.start_jenkins_job.call_count,
                         0)


class LaunchpadTriggerAutolanding(TestLaunchpadTrigger):
    def setUp(self):
        super(LaunchpadTriggerAutolanding, self).setUp()
        self.launchpadTrigger.user_allowed_to_trigger_jobs = MagicMock(
            return_value=True)
        self.launchpadTrigger.testing_in_progress = MagicMock(
            return_value=False)
        self.launchpadTrigger.unapproved_prerequisite_exists = MagicMock(
            return_value=False)
        self.launchpadTrigger.is_commit_message_set = MagicMock(
            return_value=True)
        self.launchpadTrigger.start_jenkins_job = MagicMock()

    def test_trigger_al_with_unapproved_prerequisite(self):
        self.launchpadTrigger.trigger_al_build(
            [MagicMock(), MagicMock(), MagicMock()], 'job', 'url', None)
        self.assertEqual(self.launchpadTrigger.start_jenkins_job.call_count,
                         3)


class TestGetMergeProposalsWithInvalidBranch(TestLaunchpadTrigger):
    def setUp(self):
        super(TestGetMergeProposalsWithInvalidBranch, self).setUp()
        self.launchpadTrigger._launchpad = MagicMock()
        self.launchpadTrigger._launchpad.branches.getByUrl = MagicMock(
            return_value=None)

    def test_get_merge_proposals_with_invalid_branch(self):
        self.assertEqual(self.launchpadTrigger.get_merge_proposals('', ''),
                         [])


class LaunchpadTriggerGetExecutedTestRuns(TestLaunchpadTrigger):

    executed_test_runs_message = dedent(
        """
     Executed test runs:
         FAILURE: http://{hostname}/job/job-multiconfig/./""" +
        """distribution=raring,flavor=amd64/3/console
         FAILURE: http://{hostname}/job/job-multiconfig/./""" +
        """distribution=raring,flavor=i386/3/console
         SUCCESS: http://{hostname}/job/job-freestyle/7
             deb: http://{hostname}/job/job-freestyle/7/artifact/work/""" +
        """output/*zip*/output.zip
         UNSTABLE: http://{hostname}/job/freestyle-downstream/12
     Coverity artifacts:
         http://{hostname}/job/job-multiconfig/./distribution=""" +
        """raring,flavor=amd64/3/artifact/results/coverity/CID_10895.html
         http://{hostname}/job/job-multiconfig/./distribution=""" +
        """raring,flavor=amd64/3/artifact/results/coverity/CID_10896.html
         http://{hostname}/job/job-multiconfig/./distribution=""" +
        """raring,flavor=i386/3/artifact/results/coverity/CID_10895.html
         http://{hostname}/job/job-multiconfig/./distribution=""" +
        """raring,flavor=i386/3/artifact/results/coverity/""" +
        """CID_10896.html""")

    def _get_executed_test_runs_message(self, expected_jenkins_hostname):
        return self.executed_test_runs_message.format(
            hostname=expected_jenkins_hostname)

    def setUp(self):
        super(LaunchpadTriggerGetExecutedTestRuns, self).setUp()
        jenkins = JenkinsJSONData()
        self.build_url = jenkins.build_url
        self.launchpadTrigger._get_json_jenkins = MagicMock(
            return_value=jenkins)

    def test_get_executed_test_runs_message(self):
        ret = self.launchpadTrigger.get_executed_test_runs_message(
            self.build_url)
        self.assertEqual(ret,
                         self._get_executed_test_runs_message('jenkins:8080'))

    def test_format_message_for_mp_update(self):
        test_runs = self._get_executed_test_runs_message('jenkins:8080')
        expected = dedent("""\
     mymessage
     http://jenkins:8080/job/my-job/3/{test_runs}

     Click here to trigger a rebuild:
     http://jenkins:8080/job/my-job/3/rebuild
             """).format(test_runs=test_runs)
        actual = self.launchpadTrigger.format_message_for_mp_update(
            self.build_url,
            "mymessage\n")
        self.assertEquals(actual, expected)

    def test_format_message_with_empty_message(self):
        test_runs = self._get_executed_test_runs_message('jenkins:8080')
        expected = dedent("""\
     http://jenkins:8080/job/my-job/3/{test_runs}

     Click here to trigger a rebuild:
     http://jenkins:8080/job/my-job/3/rebuild
              """).format(test_runs=test_runs)

        actual = self.launchpadTrigger.format_message_for_mp_update(
            self.build_url)
        self.assertEquals(actual, expected)

    def test_format_message_with_publishing_job(self):
        test_runs = self._get_executed_test_runs_message(
            'jenkins.qa.ubuntu.com')
        expected = dedent("""\
     http://jenkins.qa.ubuntu.com/job/my-job/3/{test_runs}

     Click here to trigger a rebuild:
     http://jenkins:8080/job/my-job/3/rebuild
              """).format(test_runs=test_runs)

        with patch("launchpad.jenkinsutils.is_job_published",
                   new=lambda x: True):
            actual = self.launchpadTrigger.format_message_for_mp_update(
                self.build_url)
            self.assertEquals(actual, expected)


class LaunchpadTriggerTestGetMergeProposals(TestLaunchpadTrigger):
    def setUp(self):
        super(LaunchpadTriggerTestGetMergeProposals, self).setUp()
        self.launchpadTrigger._launchpad = MagicMock()
        branch = MagicMock()
        branch.landing_candidates = []
        for i in range(1, 10):
            mp = MagicMock()
            mp.queue_status = i
            branch.landing_candidates.append(mp)
        self.launchpadTrigger._launchpad.branches.getByUrl = MagicMock(
            return_value=branch)

    def test_get_merge_proposals_odd(self):
        mps = self.launchpadTrigger.get_merge_proposals('', [1, 3, 5, 7])
        self.assertEqual(len(mps), 4)

    def test_get_merge_proposal_two(self):
        mps = self.launchpadTrigger.get_merge_proposals('', [2])
        self.assertEqual(len(mps), 1)
        self.assertEqual(mps[0].queue_status, 2)


class LaunchpadTriggerIsCommitMessageSet(TestLaunchpadTrigger):
    def setUp(self):
        super(LaunchpadTriggerIsCommitMessageSet, self).setUp()

    def _setup_json_data(self, json_data):
        jenkins = MagicMock()
        jenkins.get_json_data = MagicMock(return_value=json_data)
        self.launchpadTrigger._get_json_jenkins = MagicMock(
            return_value=jenkins)

    def _setup_get_commit_message(self, return_value):
        self.launchpadTrigger.get_commit_message = MagicMock(
            return_value=return_value)

    def test_use_description_but_still_no_commit_message(self):
        json_data = {'property': ['dummy', {'parameterDefinitions': [{
            'name': 'use_description_for_commit',
            'defaultParameterValue': {'value': True}}]}]}
        self._setup_json_data(json_data)
        self._setup_get_commit_message(False)
        ret = self.launchpadTrigger.is_commit_message_set(None, 'job', 'url')
        self.launchpadTrigger.get_commit_message.assert_called_once_with(None,
                                                                         True)
        self.assertFalse(ret)

    def test_no_use_description_but_commit_message_set(self):
        json_data = {'property': ['dummy', {'parameterDefinitions': [{
            'name': 'use_description_for_commit',
            'defaultParameterValue': {'value': False}}]}]}
        self._setup_json_data(json_data)
        self._setup_get_commit_message(True)
        ret = self.launchpadTrigger.is_commit_message_set(None, 'job', 'url')
        self.launchpadTrigger.get_commit_message.assert_called_once_with(None,
                                                                         False)
        self.assertTrue(ret)

    def test_use_description_flag_not_available(self):
        json_data = {}
        self._setup_json_data(json_data)
        self._setup_get_commit_message(True)
        ret = self.launchpadTrigger.is_commit_message_set(None, 'job', 'url')
        self.launchpadTrigger.get_commit_message.assert_called_once_with(None,
                                                                         False)
        self.assertTrue(ret)
