#!/usr/bin/env python

import sys
from argparse import ArgumentParser
from launchpad import LaunchpadTrigger
from settings import LOCK_NAME, logger
from socketlock import SocketLock
import re


def main():
    parser = ArgumentParser(description="Vote on a Launchpad merge proposal.")
    parser.add_argument('-s', '--status')
    parser.add_argument('-u', '--build-url', required=True,
                        help="URL of the Jenkins job")
    parser.add_argument('-r', '--revision', required=True,
                        help="merge proposal candidate revision")
    parser.add_argument('-b', '--branch',
                        help="Launchpad branch proposed for merging")
    parser.add_argument('-p', '--merge-proposal', required=True,
                        help="URL of the merge proposal to update")
    args = vars(parser.parse_args())
    launchpad = LaunchpadTrigger()
    if not args['branch']:
        # if branch was not specified try to get it from merge proposal
        args['branch'] = launchpad.get_branch_from_mp(args['merge_proposal'])
        if not args['branch']:
            parser.error('Error: unable to get branch from merge proposal.\n' +
                         'Please specify manually using -- branch.')
    mp = launchpad.get_merge_proposal(args['branch'], args['merge_proposal'])
    if not mp:
        parser.error('merge proposal related to this branch was not found')

    # this is the status from tests
    overal_status = args['status']
    # by default reason is empty as it is usually just a failed build
    reason = ''

    #override the status to FAILED and set reason if no commit message is set
    match = re.match('^(.*)/job/([^/]*)/.*$', args['build_url'])
    if match:
        jenkins_url = match.group(1)
        jenkins_job = match.group(2)
        if not launchpad.is_commit_message_set(mp, jenkins_job, jenkins_url):
            overal_status = 'FAILED'
            reason = "No commit message was specified in the merge " + \
                     "proposal. Click on the following link and set the " + \
                     "commit message (if you want a jenkins rebuild you " +\
                     "need to trigger it yourself):\n" + \
                     args['merge_proposal'] + "/+edit-commit-message\n"
            logger.debug('Commit message not set. Failing CI.')
    else:
        logger.debug('Unable to get job name from build_url.' +
                     'Skipping check for empty commit message')

    if overal_status == 'PASSED':
        launchpad.approve_mp(mp, args['revision'], args['build_url'])
    else:  # status == False corresponds to NOT 'PASSED'
        launchpad.disapprove_mp(mp,
                                args['revision'],
                                args['build_url'],
                                reason)
    sys.exit(0)


if __name__ == "__main__":
    # this is here because launchpadlib is not thread safe (yet)
    # see also lp:459418 and lp:1025153
    logger.debug('Going to acquire launchpad lock for voteOnMergeProposal')
    with SocketLock(LOCK_NAME):
        logger.debug('Lock acquired for voteOnMergeProposal')
        sys.exit(main())
